/*
 * MSC (MMC/SD controller) peripheral support.
 *
 * Copyright (C) 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/re/c/dma_space.h>
#include <l4/sys/types.h>
#include <stdint.h>



#ifdef __cplusplus

#include <l4/devices/hw_mmio_register_block.h>



// MMC/SD structures.

struct CID
{
  uint16_t month:4, year:8, reserved:4;
  uint32_t serial;
  uint8_t revision;
  char name[5];
  uint16_t oem;
  uint8_t manufacturer;
} __attribute__((packed));

struct CSD
{
  uint8_t reserved0:2, format:2, temp_write_prot:1, perm_write_prot:1, copy:1, format_group:1;
  uint16_t reserved1:5, write_block_partial:1, write_blocklen:4, write_time_factor:3,
           reserved2:2, write_prot_group_enable:1;
  uint64_t write_prot_group_size:7, erase_sector_size:7, erase_single_block_enable:1,
           device_size_multiplier:3, max_write_current_max:3, max_write_current_min:3,
           max_read_current_max:3, max_read_current_min:3, device_size:12,
           reserved3:2, dsr_implemented:1, read_block_misalign:1, write_block_misalign:1,
           read_block_partial:1, read_blocklen:4, card_command_classes:12;
  uint8_t tran_speed, data_read_access_time_2, data_read_access_time_1,
          reserved4:6, csd:2;
} __attribute__((packed));



// MMC/SD controller channel.

class Msc_channel
{
protected:
  l4_addr_t _msc_start;
  Hw::Register_block<32> _regs;
  l4_cap_idx_t _irq;

  // Support eight CID/CSD entries of 120 bits and RCA entries of 16 bits.

  struct CID _cid[8];
  struct CSD _csd[8];
  uint16_t _rca[8], _current_rca;
  uint8_t _bus_width[8], _current_bus_width;
  uint8_t _cards;

  // Utility methods.

  uint32_t get_field(uint32_t reg, uint32_t mask, uint8_t shift);
  void set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value);

  // Low-level operations.

  void ack_irq(uint32_t flags);
  void unmask_irq(uint32_t flags);
  void reset();
  void start_clock();
  void stop_clock();

  // Command properties.

  bool command_will_write(uint8_t index);
  bool command_with_data(uint8_t index);
  bool command_uses_busy(uint8_t index);
  uint8_t get_response_format(uint8_t index);

  bool app_command_will_write(uint8_t index);
  bool app_command_with_data(uint8_t index);
  bool app_command_uses_busy(uint8_t index);
  uint8_t get_app_response_format(uint8_t index);

  // Command initiation.

  bool send_app_command(uint8_t index, uint32_t arg);
  bool send_command(uint8_t index, uint32_t arg);
  bool send_command(uint8_t index, uint32_t arg, uint8_t response_format,
                    bool data, bool write, bool busy);

  // Response handling.

  bool have_response();
  void read_response(uint16_t *buffer, uint8_t units);
  bool wait_for_irq(uint32_t flags);
  bool wait_for_irq(uint32_t flags, unsigned int timeout);

  // Initialisation operations.

  bool check_sd();
  void init_sdio();
  void init_sdmem();
  void init_mmc();
  void identify_cards();
  void query_cards();

  // Transfer operations.

  uint32_t recv_data(l4re_dma_space_dma_addr_t paddr, uint32_t count);
  uint32_t send_data(l4re_dma_space_dma_addr_t paddr, uint32_t count);

  virtual uint32_t transfer(l4re_dma_space_dma_addr_t from_paddr,
                            l4re_dma_space_dma_addr_t to_paddr,
                            bool recv, uint32_t count) = 0;

public:
  explicit Msc_channel(l4_addr_t msc_start, l4_addr_t start, l4_cap_idx_t irq);

  virtual ~Msc_channel();

  void enable();

  uint32_t get_status();

  uint32_t read_block(uint8_t card, l4re_dma_space_dma_addr_t paddr);
};



// MMC/SD controller device control.

class Msc_chip
{
protected:
  l4_addr_t _msc_start, _start, _end;

  virtual unsigned int num_channels() = 0;

public:
  explicit Msc_chip(l4_addr_t msc_start, l4_addr_t start, l4_addr_t end);
};

#endif /* __cplusplus */
